
local _,rematch = ...
Rematch = rematch
local frame = RematchFrame -- the primary Rematch window
local incFrame = RematchIncomingFrame -- the window that pops up when you receive a sent team
local confirmFrame = RematchConfirmationFrame
local queueFrame = RematchLevelingQueueFrame

RematchSaved = {} -- saved teams SavedVariable
-- teams are indexed by name, either the npc name or a user-chosen name:
-- ["Name"] = { {pet1,a1,a2,a3},{pet2,a1,a2,a3},{pet3,a1,a2,a3} }
-- if an npcID is known:
-- ["Name"] = { {pet1,a1,a2,a3},{pet2,a1,a2,a3},{pet3,a1,a2,a3},npcID }
-- Pets can be species IDs (numeric), usually for pets received or imported.
-- When sending pets, they need to be sent as species IDs

RematchSettings = {} -- savedvariables settings

rematch.savedTeams = nil -- will be RematchSaved at PLAYER_LOGIN
rematch.settings = nil -- will be RematchSettings at PLAYER_LOGIN

rematch.selectedTeam = nil -- team selected in drawer list
rematch.targetedNpcID = nil -- npcID of targeted unit
rematch.targetedName = nil -- name of targeted unit

rematch.showDrawer = nil -- whether team/options are displayed in a longer frame
rematch.shareMode = nil -- 1=export, 2=import, 3=send

rematch.teamList = {} -- numerically-indexed list of teams (just name or npcID, index to saved) for the list
rematch.abilityList = {} -- reusable table for GetPetAbilityList
rematch.levelList = {} -- reusable table for GetPetABilityList
rematch.info = {} -- reusable generic table (to reduce table/garbage creation)
rematch.attackTypes = {} -- ability types that can attack, indexed by type, for generating hints
rematch.incomingTeam = {} -- used for teams sent by another user

-- attack types (index) are {strong vs type,weak vs type}
-- ie flying(3rd) is strong vs aquatic(9), weak vs dragonkin(2)
rematch.attackHints = { {2,8},{6,4},{9,2},{1,9},{4,1},{3,10},{10,5},{5,3},{7,6},{8,7} }

rematch.optionsText = {
	AutoShow={"Auto show","When targeting something with saved pets, and the pets are not already loaded, show the Rematch window and close it when pets load."},
	AutoLoad={"Auto load","When your target has saved pets not already loaded, load the pets immediately. Note: if you target by right-clicking, it may be too late to load pets!"},
	AutoMouseover={"On mouseover","Instead of waiting to target, auto-load pets when you mouseover something that has a saved team not already loaded."},
	KeepSummoned={"Keep companion","After a team is loaded, summon back the companion that was at your side before the load."},
	KeepLeveling={"Keep leveling pet","If one of your current pets is a leveling pet, loading a team will not swap out that pet. Instead, only the first two saved pets will load alongside it."},
	DisableShare={"Disable sharing","Disable the Send button and also block any incoming pets sent by others. Import and Export still work."},
	LargeWindow={"Larger window","Make the Rematch window larger for easier viewing."},
	LockWindow={"Lock window","Lock the Rematch window so it can't be dismissed with the Escape key or moved unless Shift is held."},
	AutoConfirm={"Auto confirm","Don't ask for confirmation when saving over an existing team or deleting one."},
	BestOfSpecies={"Auto upgrade","When learning a pet, automatically upgrade any lesser versions of the pet in your saved teams."},
}

rematch.shareText = {
	{"Exporting","The text below describes this team in a format usable by other users. Hit Ctrl+C to copy it to the clipboard. Ctrl+V to paste it into an email, forum or elsewhere."},
	{"Importing","If you have a team copied to the clipboard, paste it here with Ctrl+V. You can review the team above before hitting Enter or clicking Save to save the team."},
	{"Sending","Who would you like to send this team to?"},
}

rematch.levelingIcon = "Interface\\Icons\\PetJournalPortrait" -- icon to use for a leveling pet

BINDING_HEADER_REMATCH = "Rematch"

rematch.oldDropDownInit = nil -- will be the pet menu initialization function (PetJournalPetOptionsMenu or PetJournalEnhancedPetMenu)

function rematch:OnEvent(event,...)
	if rematch[event] then
		rematch[event](rematch,...)
	end
end
frame:RegisterEvent("PLAYER_LOGIN")

function rematch:PLAYER_LOGIN()
	rematch.savedTeams = RematchSaved
	rematch.settings = RematchSettings
	self:UpgradeSavedVars()
	-- if savedvar exists, position window
	if rematch.settings.XPos then
		frame:SetScale(rematch.settings.LargeWindow and 1.25 or 1)
		frame:ClearAllPoints()
		frame:SetPoint("BOTTOMLEFT",rematch.settings.XPos,rematch.settings.YPos)
	end
	rematch:BuildUI()
	rematch:OrderTeamList()

	frame:RegisterEvent("PLAYER_TARGET_CHANGED")
	frame:RegisterEvent("PLAYER_REGEN_DISABLED") -- to close rematch when entering combat
	frame:RegisterEvent("PLAYER_REGEN_ENABLED") -- to resume target watching when leaving combat
	RegisterAddonMessagePrefix("Rematch") -- for listening to our own CHAT_MSG_ADDON
	frame:RegisterEvent("CHAT_MSG_ADDON")

	frame:RegisterEvent("UPDATE_SUMMONPETS_ACTION")

	-- pet journal stuff needs to wait until it loads
	if IsAddOnLoaded("Blizzard_PetJournal") then
		rematch:ADDON_LOADED("Blizzard_PetJournal") -- another addon forced a load, run the event manually
	else
		frame:RegisterEvent("ADDON_LOADED")
	end

	hooksecurefunc(C_PetJournal,"CagePetByID",rematch.CagePetByID)

	SlashCmdList["REMATCH"] = function(msg)
		if msg and msg:lower()=="migrate" then
			rematch:MigratePBT()
		else
			frame:Toggle()
		end
	end
	SLASH_REMATCH1 = "/rematch"
end

-- when pet journal is loaded/shown, we create/anchor a rematch button to it
function rematch:ADDON_LOADED(addon)
	if addon=="Blizzard_PetJournal" then
		frame:UnregisterEvent("ADDON_LOADED")

		-- Rematch button the pet journal
		local button = CreateFrame("Button","RematchPetJournalButton",PetJournal,"MagicButtonTemplate")
		button:SetSize(140,22)
		button:SetText("Rematch")
		button:SetScript("OnClick",frame.Toggle)
		button:SetScript("OnEnter",function(self)
			GameTooltip:SetOwner(self, "ANCHOR_RIGHT")
			GameTooltip:SetText("Rematch",1,1,1)
			GameTooltip:AddLine("Toggles the Rematch window to manage battle pet teams.",nil,nil,nil,true)
			GameTooltip:Show()
		end)
		button:SetScript("OnLeave",function(self)
			GameTooltip:Hide()
		end)
		-- when pet journal is shown, adjust position of Rematch button based on existence of non-default buttons
		PetJournal:HookScript("OnShow",function(self)
			local ourButton = RematchPetJournalButton
			local leftJournal = PetJournal:GetLeft()
			local midJournal = PetJournal:GetWidth()/2 + leftJournal
			local topJournalButtons = PetJournal:GetBottom()+30
			local leftUsed,rightUsed
			for _,button in pairs({PetJournal:GetChildren()}) do
			  local topButton = button:GetTop()
			  if topButton and topButton<topJournalButtons and button~=PetJournalFindBattle and button~=PetJournalSummonButton and button~=ourButton then
					if button:GetLeft() < midJournal then -- something is in the bottomleft next to default summon button
						leftUsed = true
					else -- something is in the bottomright next to default find battle button
						rightUsed = true
					end
			  end
			end
			ourButton:ClearAllPoints()
			if not rightUsed then
				ourButton:SetPoint("RIGHT",PetJournalFindBattle,"LEFT")
			elseif not leftUsed then
				ourButton:SetPoint("LEFT",PetJournalSummonButton,"RIGHT")
			end
			ourButton:SetShown(not (rightUsed and leftUsed))
		end)

		-- hook the pet menu drop-down menu to rematch.NewDropDownMenu after giving time for all LoadWith to get done
		rematch:StartTimer("DropDownInit",.5,function()
			local parent = PetJournalEnhancedPetMenu or PetJournalPetOptionsMenu
			rematch.oldDropDownInit = parent.initialize
			parent.initialize = rematch.NewPetMenuDropDownInit
		end) -- come back in .5 seconds to look for them

	end
end

function rematch:UpdateTargetedInfo()
	rematch.targetedName,rematch.targetedNpcID = rematch:GetUnitNameandID("target")
	rematch.selectedTeam = nil -- ordinarily, clear selectedTeam when changing targets
	if rematch.savedTeams[rematch.targetedName] and (not rematch.savedTeams[rematch.targetedName][4] or rematch.savedTeams[rematch.targetedName][4]==rematch.targetedNpcID) then
		-- but if we have an npc targeted with a matching name, set selectedTeam to that target
		-- note this can set selectedTeam to named players with matching team names (which is ok)
		rematch.selectedTeam = rematch.targetedName
	end
end

function rematch:UPDATE_MOUSEOVER_UNIT(...)
	if not InCombatLockdown() then
		local mouseoverName,mouseoverNpcID = rematch:GetUnitNameandID("mouseover")
		local team = rematch.savedTeams[mouseoverName]
		if team and (not team[4] or team[4]==mouseoverNpcID) and rematch:PetsNeedLoading(mouseoverName) then
			rematch:LoadPets(mouseoverName)
		end
	end
end

function rematch:PLAYER_TARGET_CHANGED()
	rematch:UpdateTargetedInfo()
	if rematch.selectedTeam and (rematch.settings.AutoShow or rematch.settings.AutoLoad) then
		local petsNeedLoading = rematch:PetsNeedLoading(rematch.selectedTeam)
		if petsNeedLoading then
			if rematch.settings.AutoShow then
				frame:Show()
				rematch.selectedTeam = rematch.targetedName
			end
			if rematch.settings.AutoLoad then
				rematch:LoadPets(rematch.selectedTeam)
			end
		end
	end

	if frame:IsVisible() then
		if frame.drawer:IsVisible() then
			rematch:UpdateTeamList()
		end
		rematch:UpdateSavedPets()
	end
end

function rematch:PLAYER_REGEN_DISABLED()
	frame:UnregisterEvent("PLAYER_TARGET_CHANGED")
	if frame:IsVisible() then
		frame.returnAfterCombat = rematch.settings.LockWindow -- come back after combat if LockWindow set
		frame:Hide()
	end
end

function rematch:PLAYER_REGEN_ENABLED()
	frame:RegisterEvent("PLAYER_TARGET_CHANGED")
	if frame.returnAfterCombat then
		frame:Show()
		frame.returnAfterCombat = nil
	end
	if rematch.preLoadCompanion then
		rematch:RestoreCompanion()
	end
	if rematch.queueNeedsProcessed then
		rematch:ProcessLevelingQueue()
		rematch.queueNeedsProcessed = nil
	end
end

-- registered if a queue tries to process during a pet battle
function rematch:PET_BATTLE_OVER()
	frame:UnregisterEvent("PET_BATTLE_OVER")
	rematch:ProcessLevelingQueue()
end

-- these events are only registered while the rematch window is up

-- fired when pets are healed, resurrected, gains xp, etc
function rematch:UPDATE_SUMMONPETS_ACTION()
	local oldPetID = rematch:GetLevelingPetID()
	rematch:ProcessLevelingQueue()
	rematch:UpdateCurrentPets()
	local newPetID = rematch:GetLevelingPetID()
	if oldPetID and newPetID and oldPetID~=newPetID then -- if leveling pet changes, toast new leveling pet
		rematch:ToastNextLevelingPet(newPetID)
	end
end

-- peculiar quirk with default UI. loadouts clear on zoning but reappear soon after
function rematch:PLAYER_ENTERING_WORLD()
	if not C_PetJournal.GetPetLoadOutInfo(1) then -- if pets ran away while zoning
		rematch:StartTimer("PetsRanAway",.5,rematch.PLAYER_ENTERING_WORLD) -- come back in .5 seconds to look for them
	else
		rematch:UpdateCurrentPets()
	end
end

-- track when Revive Battle Pets & Battle Pet Bandage is used
function rematch:UNIT_SPELLCAST_SUCCEEDED(unit,_,_,_,spellID)
	-- 125439 is revive pet spell
	if unit=="player" and spellID==125439 then -- revive battle pet
		Rematch:StartTimer("UpdateHealButtons",0.1,rematch.UpdateHealButtons)
	end
end

-- can't just track when bandages used, also need to track when bandages received
-- this is only registered while the rematch window is visible
function rematch:BAG_UPDATE()
	Rematch:StartTimer("UpdateHealbuttons",0.1,rematch.UpdateHealButtons)
end

function rematch:UpdateHealButtons()
	frame.petheal.cooldown:SetCooldown(GetSpellCooldown(125439))
	local count = GetItemCount(86143)
	if count==0 then
		frame.petbandage.icon:SetVertexColor(.5,.5,.5)
	else
		frame.petbandage.icon:SetVertexColor(1,1,1)
	end
	frame.petbandage.count:SetText(GetItemCount(86143))
end

function rematch:HealButtonOnEnter()
	GameTooltip:SetOwner(self,"ANCHOR_RIGHT")
	if self:GetAttribute("type")=="spell" then
		GameTooltip:SetSpellByID(125439)
	else
		GameTooltip:SetItemByID(86143)
	end
	GameTooltip:Show()
end

-- Upgrade Rematch 1.x savedvar for 2.x: RematchSaved for teams, RematchSettings for settings
function rematch:UpgradeSavedVars()
	-- from 1.x to 2.0.3: move settings out to a separate variable
	if rematch.savedTeams.Settings then
		for k,v in pairs(rematch.savedTeams) do
			if k=="Settings" then
				for x,y in pairs(v) do
					rematch.settings[x]=y
				end
			end
		end
		rematch.savedTeams.Settings = nil
	end
	-- 2.0.6: KeepFirst depreciated, replaced by Keep[1] Keep[2] and Keep[3]
	if rematch.settings.KeepFirst then
		if rematch.settings.Keep then
			rematch.settings.Keep[1] = rematch.settings.KeepFirst
		else
			rematch.settings.Keep = {rematch.settings.KeepFirst}
		end
		rematch.settings.KeepFirst = nil
	end
	-- 2.1.3: Create LevelingQueue if one doesn't exist
	if not rematch.settings.LevelingQueue then
		rematch.settings.LevelingQueue = {rematch.settings.LevelingPetID}
	end
end

-- migrate any teams saved in Pet Battle Teams, whose name doesn't already exist, into Rematch
function rematch:MigratePBT()
	if IsAddOnLoaded("PetBattleTeams") and PetBattleTeamsDB and PetBattleTeamsDB.namespaces and PetBattleTeamsDB.namespaces.TeamManager then
		local PBTsaved = PetBattleTeamsDB.namespaces.TeamManager.global.teams
		for team=1,#PBTsaved do
			local teamName = PBTsaved[team].name or "Team: "..team
			if not rematch.savedTeams[teamName] then -- only migrate teams that don't already exist
				rematch.savedTeams[teamName] = {{},{},{}}
				for pet=1,3 do
					local petID = PBTsaved[team][pet].petID
					if petID then
						petID = petID:gsub("%x",function(c) return c:upper() end)
						rematch.savedTeams[teamName][pet][1] = petID
						for ability=1,3 do
							rematch.savedTeams[teamName][pet][ability+1] = PBTsaved[team][pet].abilities[ability]
						end
					end
				end
				print("\124cffffd200",teamName,"\124cffffffffmigrated to Rematch")
			else
				print("\124cffffd200",teamName,"\124cffffffffwas not migrated into Rematch. A team of that name already exists.")
			end
		end
		if frame.drawer:IsVisible() then
			rematch:OrderTeamList()
			rematch:UpdateTeamList()
		end
	end
end

-- this is for a secure hook to C_PetJournal.CagePetByID: when a pet is caged, go through all teams
-- and replace the petID being caged with its speciesID.
function rematch.CagePetByID(petID)
	if petID then
		rematch:ClearSelected()
		local speciesID = C_PetJournal.GetPetInfoByPetID(petID)
		if speciesID then
			for teamName,team in pairs(rematch.savedTeams) do
				for i=1,3 do
					if team[i][1]==petID then
						team[i][1] = speciesID
					end
				end
			end
		end
	end
end

function frame:Toggle()
	if InCombatLockdown() then
		print("\124cffffff00You're in combat. Blizzard has restrictions on what we can do with pets during combat. Try again when you're out of combat. Sorry!")
	else
		frame:SetShown(not frame:IsVisible())
	end
end

-- finish UI initialization bits not already covered in the xml
function rematch:BuildUI()
	frame.currentHeader.text:SetText("Current Battle Pets")
	frame.savedHeader.text:SetText("Saved for <no target>")
	frame.saveAs.header.text:SetText("Save Team As...")
	frame.options.header.text:SetText("Options")
	frame.options.version:SetText("Rematch version "..GetAddOnMetadata("Rematch","Version"))
	local modalLevel = frame.drawer.teams:GetFrameLevel()+4
	frame.saveAs:SetFrameLevel(modalLevel)
	frame.options:SetFrameLevel(modalLevel)
	frame.share:SetFrameLevel(modalLevel)
	frame.confirmBlackout:SetFrameLevel(modalLevel+2)

	-- current pet slots+health bars and their abilities
	frame.current = {} -- 3 current pet slots
	for i=1,3 do
		frame.current[i] = CreateFrame("Button",nil,frame,"RematchLargeSlotTemplate")
		local button = frame.current[i]
		button:SetPoint("TOPLEFT",(i-1)*78+10,-20)
		button:SetID(i)
		button.set = 1 -- set 1 is current pets
		frame.current[i].ability = {} -- 3 abilities of each current pet
		for j=1,3 do
			button.ability[j] = CreateFrame("Button",nil,frame.current[i],"RematchSmallSlotTemplate")
			button.ability[j]:SetPoint("TOPLEFT",frame.current[i],"TOPRIGHT",2,(j-1)*-25)
			button.ability[j]:SetSize(24,24)
		end
		button:RegisterForDrag("LeftButton")
		button:RegisterForClicks("LeftButtonUp","RightButtonUp")
	end

	-- saved pet slots
	frame.saved = {}
	for i=1,3 do
		frame.saved[i] = CreateFrame("Button",nil,frame,"RematchMediumSlotTemplate")
		local button = frame.saved[i]
		button:SetPoint("TOPLEFT",(i-1)*66+10,-121)
		button.set = 2 -- set 2 is saved pets
		button:SetHighlightTexture(nil)
		frame.saved[i].ability = {}
		for j=1,3 do
			button.ability[j] = CreateFrame("Button",nil,frame.saved[i],"RematchTinySlotTemplate")
			button.ability[j]:SetPoint("TOPLEFT",frame.saved[i],"TOPRIGHT",2,(j-1)*-16+5)
			button.ability[j]:SetHighlightTexture(nil)
		end
	end

	frame.drawer.leveling.set = 3 -- set 3 is leveling pet
	frame.drawer.leveling:RegisterForClicks("LeftButtonUp","RightButtonUp")
	frame.drawer.leveling:RegisterForDrag("LeftButton")

	-- assign icons for Revive Battle Pets (125439) and Battle Pet Bandage (86143)
	frame.petheal.icon:SetTexture((GetSpellTexture(125439)))
	frame.petbandage.icon:SetTexture(GetItemIcon(86143))

	-- dropdown menu for various pet options
	rematch.dropDownFrame = CreateFrame("Frame","RematchSharedDropDown",frame,"UIDropDownMenuTemplate")
	rematch.dropDownMenu = {} -- contents of the dropdown menu
	rematch.dropDownPetID = nil -- will be petID for the dropdown func to act upon

	-- go through each checkbutton setting, set its text and run its reflect
	for setting,opt in pairs(rematch.optionsText) do
		frame.options.checks[setting].text:SetText(opt[1])
		frame.options.checks[setting].tooltip = opt[2]
		frame.options.checks[setting].setting = setting
		if rematch.ReflectSetting[setting] then
			rematch.ReflectSetting[setting]()
		end
	end

	local scrollFrame,scrollBar

	-- setup teams HybridScrollFrame (frame.drawer.teams.scrollFrame)
	scrollFrame = frame.drawer.teams.scrollFrame
	scrollFrame.update = rematch.UpdateTeamList
	scrollFrame.stepSize = 54
	HybridScrollFrame_CreateButtons(scrollFrame,"RematchTeamTemplate")
	scrollBar = scrollFrame.scrollBar
	scrollBar.doNotHide = true
	scrollBar.trackBG:SetPoint("TOPLEFT",-2,15)
	scrollBar.trackBG:SetPoint("BOTTOMRIGHT",2,-15)
	scrollBar.trackBG:SetTexture("Interface\\ChatFrame\\ChatFrameBackground")
	scrollBar.trackBG:SetGradientAlpha("HORIZONTAL",.125,.125,.125,1,.05,.05,.05,1)

	-- setup levelingQueue HybridsScrollFrame (queueFrame.scrollFrame)
	scrollFrame = queueFrame.scrollFrame
	scrollFrame.update = rematch.UpdateLevelingQueue
	scrollFrame.stepSize = 22
	HybridScrollFrame_CreateButtons(scrollFrame,"RematchLevelingQueueTemplate")
	scrollBar = scrollFrame.scrollBar
	RematchLevelingQueueFrameScrollFrameScrollBarScrollUpButton:SetSize(14,12)
	RematchLevelingQueueFrameScrollFrameScrollBarScrollDownButton:SetSize(14,12)
	scrollBar.thumbTexture:SetSize(14,18)
	scrollBar.doNotHide = true
	scrollBar.trackBG:SetPoint("TOPLEFT",-1,15)
	scrollBar.trackBG:SetPoint("BOTTOMRIGHT",0,-15)
	scrollBar.trackBG:SetTexture("Interface\\ChatFrame\\ChatFrameBackground")
	scrollBar.trackBG:SetGradientAlpha("HORIZONTAL",.125,.125,.125,1,.05,.05,.05,1)
	queueFrame.emptyCapture:SetFrameLevel(scrollFrame:GetFrameLevel()+2)

end

function rematch:PetOnEnter()
	local petID
	if self.petID then
		if self.petID~=0 then
			petID = self.petID -- this is a normal petID
		else
			petID = rematch:GetLevelingPetID() -- fetch leveling petID if there is one
		end
		rematch:ShowFloatingPetCard(petID,"BOTTOMLEFT",self.ability and self.ability[3] or self,"BOTTOMRIGHT",0,0)
	end
end

function rematch:PetOnLeave()
	rematch:HideFloatingPetCard()
end

function rematch:PetOnDragStart()
	if self.petID and (self.set==1 or self.set==3) then
		C_PetJournal.PickupPet(self.petID)
	end
end

function rematch:PetOnReceiveDrag()
	local cursorType,petID = GetCursorInfo()
	local clear
	if cursorType=="battlepet" and petID then
		if self.set==1 and self:GetID()>0 and self:GetID()<4 then
			rematch:SummonedPetMayChange()
			C_PetJournal.SetPetLoadOutInfo(self:GetID(),petID)
			if PetJournal then
				PetJournal_UpdatePetLoadOut() -- update petjournal if it's open
			end
			clear = true
		elseif self.set==3 then
			if rematch:PetCanLevel(petID) then
				rematch:SetLevelingPet(petID)
				rematch:UpdateLevelingQueue()
				clear = true
			end
		end
	end
	if clear then
		ClearCursor()
		rematch:HideFloatingPetCard()
	end
end

-- this makes the current pets glow when a pet is picked up onto the cursor
function rematch:CURSOR_UPDATE()
  local hasPet,petID = GetCursorInfo()
	hasPet = hasPet=="battlepet"
	for i=1,3 do
		frame.current[i].glow:SetShown(hasPet)
	end
	frame.drawer.leveling.glow:Hide()
	queueFrame.glow:Hide()
	if hasPet then
		if rematch:PetCanLevel(petID) then
			frame.drawer.leveling.glow:Show()
			queueFrame.glow:Show()
		end
		frame:RegisterEvent("CURSOR_UPDATE") -- this is the only place this event is registered
	else
		frame:UnregisterEvent("CURSOR_UPDATE") -- cursor clear, stop watching cursor changes
	end
end
hooksecurefunc(C_PetJournal,"PickupPet",rematch.CURSOR_UPDATE)

function rematch:PetOnClick(button)
	if GetCursorInfo()=="battlepet" then
		rematch.PetOnReceiveDrag(self) -- a battlepet is on the cursor, pretend it's a drag receive
		return
	end
	if self.petID and self.petID~=0 then
		-- shift+click to link pet (only known pets work--default can't link missing either)
		if IsModifiedClick("CHATLINK") and type(self.petID)=="string" then
			ChatEdit_InsertLink(C_PetJournal.GetBattlePetLink(self.petID))
			return
		end
		if button=="RightButton" then
			if self.set==1 then
				rematch.dropDownPetID = self.petID
				rematch:ShowDropDown(1)
				return
			end
		end
		-- otherwise bring up pet journal and select the chosen pet/species
		if not PetJournalParent then
			PetJournal_LoadUI()
		end
		if not PetJournalParent:IsShown() then
			ShowUIPanel(PetJournalParent)
		elseif PetJournalPetCard.petID==self.petID or PetJournalPetCard.speciesID==self.petID then
			HideUIPanel(PetJournalParent)
		end
		PetJournalParent_SetTab(PetJournalParent,2)
		if type(self.petID)=="string" then
			PetJournal_SelectPet(PetJournal,self.petID)
		else -- petID is a species, a missing pet
			PetJournal_SelectSpecies(PetJournal,self.petID)
		end
	end
end

function rematch:AbilityOnClick()
	if self.abilityID then
		if IsModifiedClick("CHATLINK") then
			local petID = self:GetParent().petID
			local maxHealth,power,speed,_ = 100,0,0
			if type(petID)=="string" then
				_,maxHealth,power,speed = C_PetJournal.GetPetStats(petID)
			end
			ChatEdit_InsertLink(GetBattlePetAbilityHyperlink(self.abilityID,maxHealth,power,speed))
		elseif self:GetParent().petID then
			rematch.PetOnClick(self:GetParent()) -- pretend we clicked parent pet
		end
	end
end

function rematch:VerifyButtons()
	frame.saveButton:Enable()
	frame.deleteButton:Enable()
	frame.loadButton:Disable()
	frame.drawer.sendButton:Disable()
	frame.drawer.exportButton:Disable()
	if rematch.selectedTeam then
		if frame.drawer:IsVisible() then
			if not rematch.settings.DisableShare then
				frame.drawer.sendButton:Enable()
			end
			frame.drawer.exportButton:Enable()
		end
	elseif rematch.targetedNpcID and rematch.savedTeams[rematch.targetedName] then
		-- do nothing, keep buttons enabled
	else
		frame.deleteButton:Disable()
	end
	if rematch.selectedTeam and rematch.savedTeams[rematch.selectedTeam] then
		frame.loadButton:Enable()
	end
end

-- panel buttons like close/load/save/delete
-- this is too spaghetti - each button has a name, break it out into a table like events?
function rematch:ButtonOnClick()
	if self==frame.closeButton then
		frame:Hide()
	elseif self==frame.saveButton then
		if not rematch.selectedTeam then
			if not rematch.targetedNpcID then -- save hit with no npc target (and no team selected)
				frame.drawer:Hide()
				frame.saveAs:Show()
			else -- save hit while npc targeted (and no team selected)
				rematch:SaveCurrentTeam(rematch.targetedName,rematch.targetedNpcID)
			end
		else -- hitting save with a team selected
			local npcID
			if not rematch.targetedName and rematch.savedTeams[rematch.selectedTeam][4] then
				npcID = rematch.savedTeams[rematch.selectedTeam][4] -- if nothing targeted, reuse existing npcID
			elseif rematch.targetedName==rematch.selectedTeam then
				npcID = rematch.targetedNpcID -- if something targeted, and it's same name, use targeted npcID
			end
			rematch:SaveCurrentTeam(rematch.selectedTeam,npcID)
		end
		if frame.drawer:IsVisible() then
			rematch:OrderTeamList()
			rematch:UpdateTeamList()
		end
	elseif self==frame.saveAs.cancelButton then
		frame.saveAs:Hide()
	elseif self==frame.saveAs.saveButton then
		rematch:SaveCurrentTeam(frame.saveAs.editBoxFrame.editBox:GetText())
		frame.saveAs:Hide()
	elseif self==frame.deleteButton then
		if rematch.selectedTeam then
			rematch:DeleteTeam(rematch.selectedTeam)
		end
	elseif self==frame.loadButton then
		if rematch.selectedTeam and rematch.savedTeams[rematch.selectedTeam] then
			rematch:LoadPets(rematch.selectedTeam)
		elseif rematch.targetedName and rematch.savedTeams[rematch.targetedName] then
			rematch:LoadPets(rematch.targetedName)
		end
	elseif self==frame.drawer.optionsButton then
		frame.options:Show()
	elseif self==frame.options.okayButton then
		frame.options:Hide()
	elseif self==frame.drawer.exportButton then
		if rematch.selectedTeam then
			rematch:ExportTeam()
		end
	elseif self==frame.share.cancelButton then
		frame.share:Hide()
	elseif self==frame.drawer.importButton then
		rematch:ImportTeam()
	elseif self==frame.share.saveButton then
		rematch:SaveShare()
	elseif self==frame.drawer.sendButton then
		if rematch.selectedTeam then
			rematch.shareMode = 3
			frame.share:Show()
		end
	elseif self==incFrame.cancelButton then
		incFrame:Hide()
	elseif self==incFrame.saveButton then
		rematch:SaveIncoming()
	elseif self==confirmFrame.noButton then
		confirmFrame:Hide()
	elseif self==confirmFrame.yesButton then
		confirmFrame.func()
--		rematch:SaveConfirm()
		confirmFrame:Hide()
	end
end

function rematch:SaveAsEditBoxChanged()
	if self:GetText():len()>0 then
		frame.saveAs.saveButton:Enable()
	else
		frame.saveAs.saveButton:Disable()
	end
	local exists = rematch.savedTeams[self:GetText()] and true
	frame.saveAs.warning:SetShown(exists)
	frame.saveAs.warningIcon:SetShown(exists)
end

--[[ OnShow/OnHides ]]

-- esc is heavily used in rematch to close panels. when the drawer is shown,
-- it disables esc on parent frame and enables it on drawer. when options from drawer
-- is shown, it disables esc on drawer and enables it on options.
-- frame:Hide() doesn't run an OnHide handler if the frame is already hidden
-- remember that at any time the whole frame needs hidden due to entering combat

-- being in UISpecialFrames makes a frame hide on esc key
function rematch:SetESCable(frame,enable)
	if frame=="RematchFrame" and rematch.settings.LockWindow then
		enable = false -- disable esc for main window if LockWindow set
	end
	local hasFrame = tContains(UISpecialFrames,frame)
	if enable and not hasFrame then
		tinsert(UISpecialFrames,frame)
	elseif not enable and hasFrame then
		for i=1,#UISpecialFrames do
			if UISpecialFrames[i]==frame then
				table.remove(UISpecialFrames,i)
				return
			end
		end
	end
end

function rematch:OnShow()
	if not C_PetJournal.GetPetLoadOutInfo(1) then
		-- due to lag or from summoning the window before pets load, loadout pets may not exist.
		-- in those cases come back in half a second and try again
		rematch:StartTimer("PetsRanAway",.5,rematch.OnShow)
		return
	end
	rematch:ProcessLevelingQueue()
--	rematch:SetLevelingPet() -- (rematch.settings.LevelingPetID)
--	rematch:PLAYER_TARGET_CHANGED() -- update target stuff
	rematch:UpdateTargetedInfo()
	rematch:UpdateCurrentPets()
	rematch:UpdateSavedPets()
	if frame.drawer:IsVisible() then
		rematch:UpdateTeamList()
	end
	rematch:WatchForChangingPets()
	rematch:UpdateHealButtons()
	rematch:SetESCable("RematchFrame",true)
	frame:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED")
	frame:RegisterEvent("BAG_UPDATE")
	frame:RegisterEvent("PLAYER_ENTERING_WORLD")
end

function rematch:OnHide()
	frame:UnregisterEvent("PLAYER_ENTERING_WORLD")
	frame:UnregisterEvent("BAG_UPDATE")
	frame:UnregisterEvent("UNIT_SPELLCAST_SUCCEEDED")
	frame.options:Hide()
	frame.share:Hide()
	frame.drawer:Hide()
	frame.saveAs:Hide()
	RematchFloatingPetCard:Hide()
	rematch:SetESCable("RematchFrame",false)
end

function rematch:DrawerOnShow()
	frame.saveAs:Hide()
	rematch.showDrawer = true
	rematch:OrderTeamList()
	rematch:UpdateTeamList()
	frame:SetHeight(374) -- was 342
	frame.toggle:SetNormalTexture("Interface\\ChatFrame\\UI-ChatIcon-Minimize-Up")
	frame.toggle:SetPushedTexture("Interface\\ChatFrame\\UI-ChatIcon-Minimize-Down")
	rematch:SetESCable("RematchFrame",false)
	rematch:SetESCable("RematchDrawer",true)
end

function rematch:DrawerOnHide()
	rematch.showDrawer = nil
	rematch:ClearSelected()
	rematch.settings.XPos,rematch.settings.YPos = frame:GetLeft(),frame:GetBottom()
	frame:SetHeight(196) -- was 196
	frame.toggle:SetNormalTexture("Interface\\ChatFrame\\UI-ChatIcon-Maximize-Up")
	frame.toggle:SetPushedTexture("Interface\\ChatFrame\\UI-ChatIcon-Maximize-Down")
	rematch:SetESCable("RematchDrawer",false)
	rematch:SetESCable("RematchFrame",true)
	frame:ClearAllPoints()
	frame:SetPoint("BOTTOMLEFT",rematch.settings.XPos,rematch.settings.YPos)
end

function rematch:SaveAsOnShow()
	rematch:SetESCable("RematchFrame",false)
	rematch:SetESCable("RematchSaveAs",true)
	frame.saveAs.saveButton:Disable()
	frame.savedHeader:Hide()
end

function rematch:SaveAsOnHide()
	rematch:SetESCable("RematchSaveAs",false)
	rematch:SetESCable("RematchFrame",true)
	frame.saveAs.editBoxFrame.editBox:SetText("")
	frame.savedHeader:Show()
end

function rematch:OptionsOnShow()
	rematch:SetESCable("RematchDrawer",false)
	rematch:SetESCable("RematchOptions",true)
	frame.savedHeader:Hide()
end

function rematch:OptionsOnHide()
	rematch:SetESCable("RematchOptions",false)
	rematch:SetESCable("RematchDrawer",true)
	frame.savedHeader:Show()
end

function rematch:ShareOnShow()
	rematch:SetESCable("RematchDrawer",false)
	rematch:SetESCable("RematchShare",true)
	frame.drawer.teams:Hide()
	frame.drawer.leveling.label:Hide()
	local shareMode = rematch.shareMode
	frame.share.header.text:SetText(rematch.shareText[shareMode][1])
	frame.share.descBox.text:SetText(rematch.shareText[shareMode][2])
	frame.share.cancelButton:SetText(shareMode==1 and "Ok" or "Cancel")
	frame.share.detailBox.text:SetText("")
	frame.share.saveButton:SetText(shareMode==2 and "Save" or "Send")
	frame.share.saveButton:SetShown(shareMode~=1) -- don't show save/send button for Export
	rematch:SetShareStatus() -- clear any warning
end

function rematch:ShareOnHide()
	rematch:SetESCable("RematchShare",false)
	rematch:SetESCable("RematchDrawer",true)
	if rematch.selectedTeam and rematch.selectedTeam:match("^~temp~") then
		rematch.selectedTeam = nil
	end
	rematch:UpdateSavedPets()
	rematch:OrderTeamList()
	rematch:UpdateTeamList()
	frame.drawer.leveling.label:Show()
	frame.drawer.teams:Show()
	frame:UnregisterEvent("CHAT_MSG_SYSTEM") -- in case any sending was not finished
end

function rematch:ConfirmOnShow()
	rematch:SetESCable("RematchConfirmationFrame",true)
	frame.confirmBlackout:Show()
end

function rematch:ConfirmOnHide()
	frame.confirmBlackout:Hide()
	rematch:SetESCable("RematchConfirmationFrame",false)
end

function rematch:SaveCurrentTeam(team,npcID)
--	rematch.savedTeams[team] = {}
	rematch.incomingTeam = {}
	for i=1,3 do
		local c = frame.current[i]
		if rematch:IsCurrentLevelingPet(c.petID) then
			tinsert(rematch.incomingTeam,{0,0,0,0}) -- this is a leveling pet, save it as 0s
		else
			tinsert(rematch.incomingTeam,{c.petID,c.ability[1].abilityID,c.ability[2].abilityID,c.ability[3].abilityID})
		end
	end
	if npcID then
		tinsert(rematch.incomingTeam,npcID)
	end
	rematch:SaveTeam(team,rematch.incomingTeam)
--	rematch.selectedTeam = team
--	rematch:UpdateSavedPets()
--	for i=1,3 do
--		frame.saved[i].shine:SetCooldown(0,1) -- do shiny cooldown ending bit
--	end
end

-- check buttons autoShow, autoLoad, etc
function rematch:CheckButtonOnClick()
	rematch.settings[self.setting] = self:GetChecked()
	if rematch.ReflectSetting[self.setting] then
		rematch.ReflectSetting[self.setting]()
	end
end

function rematch:CheckButtonOnEnter()
	frame.options.tooltip.text:SetText(self.tooltip)
end

-- only call from the toggle drawer button
function rematch:ToggleDrawer()
	rematch.showDrawer = not rematch.showDrawer
	rematch.settings.XPos,rematch.settings.YPos = frame:GetLeft(),frame:GetBottom()
	frame.drawer:SetShown(rematch.showDrawer)
	frame:ClearAllPoints()
	frame:SetPoint("BOTTOMLEFT",rematch.settings.XPos,rematch.settings.YPos)
end

function rematch:StartMoving()
	if not rematch.settings.LockWindow or IsShiftKeyDown() then
		frame.isMoving = true
		frame:StartMoving()
	end
end

function rematch:StopMoving()
	if frame.isMoving then
		frame.isMoving = nil
		frame:StopMovingOrSizing()
		rematch.settings.XPos,rematch.settings.YPos = frame:GetLeft(),frame:GetBottom()
	end
end

--[[ Team List ]]

function rematch:TeamFormattedProperly(team)
	if type(team)~="table" then
		return
	end
	for i=1,3 do
		if not team[i] or type(team[i])~="table" then
			return
		end
	end
	return true
end

function rematch:OrderTeamList()
	wipe(rematch.teamList)
	for k,v in pairs(rematch.savedTeams) do
		if k:match("^~temp~") or not rematch:TeamFormattedProperly(v) then
			rematch.savedTeams[k] = nil -- remove any temporary teams
		else
			tinsert(rematch.teamList,k)
		end
	end
	table.sort(rematch.teamList,function(e1,e2) return e1:lower()<e2:lower() end)
end

function rematch:UpdateTeamList()
	local numData = #rematch.teamList
	local scrollFrame = frame.drawer.teams.scrollFrame
	local offset = HybridScrollFrame_GetOffset(scrollFrame)
	local buttons = scrollFrame.buttons
	for i=1, #buttons do
		local index = i + offset
		local button = buttons[i]
		button:Hide()
		button.selectedTexture:Hide()
		if ( index <= numData) then
			local team = rematch.teamList[index]
			button.team = team
			if team==rematch.selectedTeam then
				button.selectedTexture:Show()
			end
			if rematch.savedTeams[team][4] then
				button.name:SetTextColor(1,1,1)
			else
				button.name:SetTextColor(1,.82,0)
			end
			button.name:SetText(team)
			for j=1,3 do
				local pet = rematch.savedTeams[team][j][1]
				local icon
				if pet==0 then
					icon = rematch.levelingIcon
				elseif type(pet)=="string" then -- this is an owned pet
					icon = select(9,C_PetJournal.GetPetInfoByPetID(pet))
				elseif type(pet)=="number" then -- this is a missing pet (species ID)
					icon = select(2,C_PetJournal.GetPetInfoBySpeciesID(pet))
				end
				button[tostring(j)]:SetTexture(icon or "Interface\\Buttons\\UI-PageButton-Background")
			end
			button:Show()
		end
	end
	HybridScrollFrame_Update(scrollFrame,18*numData,18)
	rematch:VerifyButtons()
end

--[[ Current pets ]]

-- updates the 3 current pets and their abilities
function rematch:UpdateCurrentPets()
	local petID
	rematch:WipePetFrames(frame.current)
	wipe(rematch.info)
	local ability = rematch.info
	for i=1,3 do
		petID,ability[1],ability[2],ability[3] = C_PetJournal.GetPetLoadOutInfo(i)
		if petID then
			frame.current[i].petID = petID
			local speciesID,_,level,xp,xpmax,_,_,_,icon = C_PetJournal.GetPetInfoByPetID(petID)
			-- update this pet's icon
		  frame.current[i].icon:SetTexture(icon)
			frame.current[i].icon:Show()
			for j=1,3 do
				frame.current[i].ability[j].abilityID = ability[j]
				-- confirm pet is high enough to have this ability
				wipe(rematch.abilityList)
				C_PetJournal.GetPetAbilityList(speciesID,rematch.abilityList,rematch.levelList)
				for k=1,#rematch.abilityList do
					if ability[j]==rematch.abilityList[k] and level>=rematch.levelList[k] then
						frame.current[i].ability[j].icon:SetTexture((select(2,C_PetJournal.GetPetAbilityInfo(ability[j]))))
						frame.current[i].ability[j].icon:Show()
						break
					end
				end
			end
			-- xp bar: update+show if pet is less than 25, hide if pet is 25
			if level<25 then
				frame.current[i].healthBG:SetPoint("TOPLEFT",frame.current[i],"BOTTOMLEFT",2,1)
				frame.current[i].xpBG:Show()
				frame.current[i].xp:SetWidth(xp>0 and 38*(xp/xpmax) or 1)
				frame.current[i].xp:Show()
				frame.current[i].level.text:SetText(level)
				frame.current[i].level:Show()
			else
				frame.current[i].healthBG:SetPoint("TOPLEFT",frame.current[i],"BOTTOMLEFT",2,-2)
				frame.current[i].xpBG:Hide()
				frame.current[i].xp:Hide()
				frame.current[i].level:Hide()
			end
			-- update hp and whether pet is dead
			local hp,hpmax = C_PetJournal.GetPetStats(frame.current[i].petID)
			if hp==0 then
				frame.current[i].dead:Show() -- dead pet :(
				frame.current[i].health:Hide()
			else
				frame.current[i].dead:Hide()
				frame.current[i].health:SetWidth(hp>0 and 38*(hp/hpmax) or 1)
				frame.current[i].health:Show()
			end
		end
	end
	rematch:UpdateLevelingBorders()
end

-- there's no event for when loadout pets change (really!) so we hooksecurefunc both
-- SetAbility and SetPetLoadOutInfo to do an update 0.1 seconds after they happen.
-- the wait is to allow multiple simultaneous changes to happen before doing an update
function rematch:WatchForChangingPets()
	if not rematch.loadoutsHooked then
		rematch.loadoutsHooked = 1
		hooksecurefunc(C_PetJournal,"SetAbility",rematch.StartPetsChanging)
		hooksecurefunc(C_PetJournal,"SetPetLoadOutInfo",rematch.StartPetsChanging)
	end
end
function rematch:StartPetsChanging()
	rematch:StartTimer("PetsChanging",0.1,rematch.UpdateCurrentPets)
end

--[[ Saving teams ]]

function rematch:TeamOnClick()
	local team = self.team
	if rematch.selectedTeam==team then
		rematch.selectedTeam = nil
	else
		rematch.selectedTeam = team
	end
	rematch:UpdateTeamList()
	rematch:UpdateSavedPets()
end

function rematch:TeamOnDoubleClick()
	rematch.selectedTeam = self.team
	rematch:UpdateTeamList()
	rematch:UpdateSavedPets()
	rematch:LoadPets(self.team)
end

-- wipes frames of any leftover pets/abilities
-- parent is likely frame.current, frame.saved or incFrame.pets
function rematch:WipePetFrames(parent)
	for i=1,3 do
		parent[i].petID = nil
		parent[i].icon:Hide()
		parent[i].leveling:Hide()
		for j=1,3 do
			parent[i].ability[j].abilityID = nil
			parent[i].ability[j].icon:Hide()
		end
	end
end

function rematch:UpdateSavedPets()
	rematch:WipePetFrames(frame.saved)
	for i=1,6 do
		frame.hints[tostring(i)]:Hide()
	end
	if rematch.selectedTeam then -- then see if a team is selected
		rematch:FillSavedFromTeam(rematch.selectedTeam)
		local actualTeam = rematch.selectedTeam:gsub("~temp~","") -- remove temporary tag if it exists
		if rematch.savedTeams[rematch.selectedTeam][4] then
			frame.savedHeader.text:SetText(format("Saved for \124cffffffff%s",actualTeam))
		else
			frame.savedHeader.text:SetText(format("Saved as \"%s\"",actualTeam))
		end
	elseif UnitExists("target") then
		if rematch.targetedNpcID then -- this is an npc
			if rematch.savedTeams[rematch.targetedName] and rematch.savedTeams[rematch.targetedName][4]==targetedNpc then
				rematch:FillSavedFromTeam(rematch.targetedName)
			end
			frame.savedHeader.text:SetText(format("Saved for \124cffffffff%s",rematch.targetedName))
		else -- this is a player, treat them just like having no target
			frame.savedHeader.text:SetText("Saved Battle Pets")
		end
	else -- no target, no saved team
		frame.savedHeader.text:SetText("Saved Battle Pets")
	end
	rematch:VerifyButtons()
end

-- returns name of the given unit and npcID if it's an npc, or nil if unit doesn't exist
function rematch:GetUnitNameandID(unit)
	if UnitExists(unit) then
		local name = UnitName(unit)
		local npcID = tonumber(UnitGUID(unit):sub(6,10),16)
		if npcID~=0 then
			return name,npcID -- this is an npc, return its name and npcID
		else
			return name -- this is a player, return its name
		end
	end
end

function rematch:DesaturateIfMissing(texture,missing)
	texture:SetDesaturated(missing)
	if missing then
		texture:SetVertexColor(.75,.75,.75)
	else
		texture:SetVertexColor(1,1,1)
	end
end

function rematch:FillSavedFromTeam(team)

	local attacks = rematch.attackTypes
	for i=1,10 do
		attacks[i]=0 -- reset counter of attack types
	end

	-- this part updates the displayed saved pets and their abilities
	for i=1,3 do
		local icon,speciesName
		local petID = rematch.savedTeams[team][i][1]
		local missing
		if petID==0 then
			icon = rematch:GetLevelingPetIcon()
			frame.saved[i].leveling:SetShown(rematch:GetLevelingPetID() and true)
		elseif type(petID)=="string" then -- petID of owned pet
			icon = select(9,C_PetJournal.GetPetInfoByPetID(petID))
		elseif type(petID)=="number" then
			speciesName,icon = C_PetJournal.GetPetInfoBySpeciesID(petID)
			if speciesName then
				local _,foundPetID = C_PetJournal.FindPetIDByName(speciesName)
				if foundPetID then
					petID = foundPetID
					rematch.savedTeams[team][i][1] = petID
				else
					missing = 1
				end
			end
		end
		if icon then
			frame.saved[i].petID = petID
			frame.saved[i].icon:SetTexture(icon)
			rematch:DesaturateIfMissing(frame.saved[i].icon,missing)
			frame.saved[i].icon:Show()
			for j=1,3 do
				if rematch.savedTeams[team][i][1+j] then
					local _,_,abilityIcon,_,_,_,abilityType,noHints = C_PetBattles.GetAbilityInfoByID(rematch.savedTeams[team][i][1+j])
					if abilityIcon then
						if not noHints and abilityType then -- if ability is an attack, increment its counter
							attacks[abilityType] = attacks[abilityType]+1
						end
						frame.saved[i].ability[j].abilityID = rematch.savedTeams[team][i][1+j]
						frame.saved[i].ability[j].icon:SetTexture(abilityIcon)
						rematch:DesaturateIfMissing(frame.saved[i].ability[j].icon,missing)
						frame.saved[i].ability[j].icon:Show()
					end
				end
			end
		end
	end

	-- the following updates the hints window to show what the team's attacks
	-- are strong against.  we'll go through the attackTypes table gathered
	-- above and multiply the values by 100, add their index, and reverse sort.
	-- ie 4x flying is 403, 1x mechanical is 110.
	for i=1,10 do
		attacks[i] = attacks[i]*100+i
	end
	table.sort(attacks,function(e1,e2) return e1>e2 end)
	-- now the most common attacks are at the top. the value%100 will be the attack type
	for i=1,10 do
		attacks[i] = attacks[i]>99 and attacks[i]%100 or nil
	end
	frame.hints["1"]:SetShown(#attacks~=0) -- show strong/weak green arrow if there are any attacks
	local centerRow
	for i=1,5 do -- only care about first 5
		if attacks[i] then
			local strong = tostring(i+1)
			frame.hints[strong]:SetTexture("Interface\\PetBattles\\PetIcon-"..PET_TYPE_SUFFIX[rematch.attackHints[attacks[i]][1]])
			frame.hints[strong]:Show()
		elseif i==3 then
			centerRow = true
		end
	end
	-- move first row to center if less than 3 attacks to display
	for i=1,3 do
		frame.hints[tostring(i)]:SetPoint("CENTER",frame.hints,centerRow and "LEFT" or "TOPLEFT",(i-1)*20+13,centerRow and 0 or -13)
	end
end

--[[ Keep Summoned ]]

-- primary for KeepSummoned, call this before pets are about to change
function rematch:SummonedPetMayChange()
	if rematch.settings.KeepSummoned then
		rematch.preLoadCompanion = rematch.preLoadCompanion or C_PetJournal.GetSummonedPetGUID()
		frame:RegisterEvent("UNIT_PET")
		rematch:StartTimer("RestoreTimeout",1,rematch.RestoreTimeout)
	end
end

function rematch:UNIT_PET()
	rematch:StopTimer("RestoreTimeout")
	frame:UnregisterEvent("UNIT_PET")
	rematch:StartTimer("RestoreCompanion",1.6,rematch.RestoreCompanion) -- wait a GCD before restoring
end

function rematch:RestoreCompanion()
	if not InCombatLockdown() then -- can't SummonPetByGUID in combat :(
		local nowSummoned = C_PetJournal.GetSummonedPetGUID()
		if not rematch.preLoadCompanion and nowSummoned then
			C_PetJournal.SummonPetByGUID(nowSummoned) -- something summoned, had nothing before
		elseif nowSummoned ~= rematch.preLoadCompanion then
			C_PetJournal.SummonPetByGUID(rematch.preLoadCompanion) -- something summoned different than before
		end
		rematch.preLoadCompanion = nil
	end
end

-- if no UNIT_PET fired, unregister it. apparently no pet was loaded
function rematch:RestoreTimeout()
	rematch.preLoadCompanion = nil
	frame:UnregisterEvent("UNIT_PET")
end

--[[ Timer Management ]]

rematch.timerFuncs = {} -- indexed by arbitrary name, the func to run when timer runs out
rematch.timerTimes = {} -- indexed by arbitrary name, the duration to run the timer
rematch.timerFrame = CreateFrame("Frame")
rematch.timerFrame:Hide()

function rematch:StartTimer(name,duration,func)
	rematch.timerFuncs[name] = func
	rematch.timerTimes[name] = duration
	rematch.timerFrame:Show()
end

function rematch:StopTimer(name)
	if name and rematch.timerTimes[name] then
		rematch.timerTimes[name] = nil
	end
end

rematch.timerFrame:SetScript("OnUpdate",function(self,elapsed)
	local tick
	local times = rematch.timerTimes
  for name in pairs(times) do
		times[name] = times[name] - elapsed
		if times[name] < 0 then
			times[name] = nil
			rematch.timerFuncs[name]()
		end
		tick = 1
	end
	if not tick then
		self:Hide()
	end
end)

--[[ Option relections ]]

rematch.ReflectSetting = {}

-- during initialization and when options are clicked, these are run.
-- note the checkbutton should be named the same as the setting

function rematch.ReflectSetting.LockWindow()
	if rematch.settings.LockWindow then
		frame.options.checks.LockWindow:SetChecked(1)
	else
		frame.options.checks.LockWindow:SetChecked(nil)
	end
	rematch:SetESCable("RematchFrame",false) -- turn it off if it's on
end

function rematch.ReflectSetting.LargeWindow()
	-- instead of a straight SetScale, we'll adjust position so it stays at the same
	-- relative position on the screen, growing out/in from the bottomleft corner
	local oldScale = frame:GetEffectiveScale()
	rematch.settings.XPos = frame:GetLeft()
	rematch.settings.YPos = frame:GetBottom()
	frame:SetScale(rematch.settings.LargeWindow and 1.25 or 1.0)
	local newScale = frame:GetEffectiveScale()
	rematch.settings.XPos = (rematch.settings.XPos*oldScale)/newScale
	rematch.settings.YPos = (rematch.settings.YPos*oldScale)/newScale
	frame:ClearAllPoints()
	frame:SetPoint("BOTTOMLEFT",rematch.settings.XPos,rematch.settings.YPos)
	frame.options.checks.LargeWindow:SetChecked(rematch.settings.LargeWindow)
	RematchFloatingPetCard:SetScale(rematch.settings.LargeWindow and 1.2 or 1)
end

function rematch.ReflectSetting.KeepSummoned()
	frame.options.checks.KeepSummoned:SetChecked(rematch.settings.KeepSummoned)
end

function rematch.ReflectSetting.KeepLeveling()
	frame.options.checks.KeepLeveling:SetChecked(rematch.settings.KeepLeveling)
end

function rematch.ReflectSetting.AutoShow()
	frame.options.checks.AutoShow:SetChecked(rematch.settings.AutoShow)
end

function rematch.ReflectSetting.AutoLoad()
	frame.options.checks.AutoLoad:SetChecked(rematch.settings.AutoLoad)
	if rematch.settings.AutoLoad then
		frame.options.checks.AutoMouseover:Enable()
		frame.options.checks.AutoMouseover.text:SetTextColor(1,.82,0)
	else
		frame.options.checks.AutoMouseover:Disable()
		frame.options.checks.AutoMouseover.text:SetTextColor(.5,.5,.5)
	end
	rematch.ReflectSetting.AutoMouseover() -- update whether UPDATE_UNIT_MOUSEOVER should be registered
end

function rematch.ReflectSetting.AutoMouseover()
	frame.options.checks.AutoMouseover:SetChecked(rematch.settings.AutoMouseover)
	if rematch.settings.AutoMouseover and rematch.settings.AutoLoad then
		frame:RegisterEvent("UPDATE_MOUSEOVER_UNIT")
	else
		frame:UnregisterEvent("UPDATE_MOUSEOVER_UNIT")
	end
end

function rematch.ReflectSetting.DisableShare()
	frame.options.checks.DisableShare:SetChecked(rematch.settings.DisableShare)
	if rematch.settings.DisableShare then
		frame.drawer.sendButton:Disable()
	else
		frame.drawer.sendButton:Enable()
	end
end

function rematch.ReflectSetting.AutoConfirm()
	frame.options.checks.AutoConfirm:SetChecked(rematch.settings.AutoConfirm)
end

function rematch.ReflectSetting.BestOfSpecies()
	frame.options.checks.BestOfSpecies:SetChecked(rematch.settings.BestOfSpecies)
	rematch:SetupBestOfSpecies()
end

function rematch:ClearSelected()
	rematch.selectedTeam = nil
	rematch.targetedName = nil
	rematch.targetedNpcID = nil
	rematch:UpdateSavedPets()
	if frame.drawer:IsVisible() then
		rematch:UpdateTeamList()
	end
end

--[[ Pet Ability Tooltips ]]

-- ability tooltips use FloatingPetBattleAbilityTooltip with a few tweaks
function rematch:AbilityOnEnter()
	rematch.borderKeys = rematch.borderKeys or {"BorderTopLeft","BorderTopRight","BorderBottomLeft","BorderBottomRight","BorderLeft","BorderRight","BorderTop","BorderBottom"}
	local petID = self:GetParent().petID
	if petID and self.abilityID then
		local _,maxHealth,power,speed
		if type(petID)=="string" then
			_,maxHealth,power,speed = C_PetJournal.GetPetStats(petID)
		else
			maxHealth,power,speed = 100,0,0 -- missing pets are weak!
		end
		FloatingPetBattleAbility_Show(self.abilityID,maxHealth,power,speed)
		FloatingPetBattleAbilityTooltip:ClearAllPoints()
		FloatingPetBattleAbilityTooltip:SetPoint("BOTTOMLEFT",self,"BOTTOMRIGHT")
		-- make border gold (Blizzard why you no use tooltip Backdrop)
		for _,key in pairs(rematch.borderKeys) do
			FloatingPetBattleAbilityTooltip[key]:SetVertexColor(.9,.738,0)
		end
		-- hide close button that can't be clicked anyway
		FloatingPetBattleAbilityTooltip.CloseButton:Hide()
		-- if LargeWindow not checked, shrink tooltip by 10%
		FloatingPetBattleAbilityTooltip:SetScale(rematch.settings.LargeWindow and 1 or .9)
		FloatingPetBattleAbilityTooltip:EnableMouse(false)
	end
end
function rematch:AbilityOnLeave()
	FloatingPetBattleAbilityTooltip:Hide()
	-- revert gold border, show close button again and restore scale
	for _,key in pairs(rematch.borderKeys) do
		FloatingPetBattleAbilityTooltip[key]:SetVertexColor(1,1,1)
	end
	FloatingPetBattleAbilityTooltip.CloseButton:Show()
	FloatingPetBattleAbilityTooltip:SetScale(1)
	FloatingPetBattleAbilityTooltip:EnableMouse(true)
end

--[[ Saving/Confirmation ]]

-- instead of disparate ways teams are saved (main save, import save, receive save), all team saves go through here
-- returns true if save happened immediately
function rematch:SaveTeam(teamName,team)
	if not rematch.settings.AutoConfirm and rematch.savedTeams[teamName] and not rematch:TeamsAreSimilar(rematch.savedTeams[teamName],team) then
		rematch:ConfirmSave(teamName,rematch.savedTeams[teamName],team)
	else
		confirmFrame.teamName = teamName
		confirmFrame.newTeam = team
		rematch:SaveConfirm()
		return true
	end
end

-- returns true if the teams are similar (same species in every slot)
function rematch:TeamsAreSimilar(team1,team2)
	for i=1,3 do
		if team1[i][1]~=team2[i][1] then
			if team1[i][1]==0 or team2[i][1]==0 then
				return false
			end
			local speciesID1 = C_PetJournal.GetPetInfoByPetID(team1[i][1])
			local speciesID2 = C_PetJournal.GetPetInfoByPetID(team2[i][1])
			if speciesID1 ~= speciesID2 then
				return false
			end
		end
	end
	return true
end

-- displays confirmation popup whether to overwrite an oldTeam with newTeam in teamName
function rematch:ConfirmSave(teamName,oldTeam,newTeam)
	rematch:PrepareConfirmation(teamName,newTeam[4])
	rematch:FillPetFrames(confirmFrame.old,oldTeam)
	rematch:FillPetFrames(confirmFrame.new,newTeam)
	confirmFrame.newTeam = newTeam
	confirmFrame.teamName = teamName
	confirmFrame.func = rematch.SaveConfirm
	confirmFrame:Show()
end

-- actual save when save is confirmed, writes the newTeam to teamName and updates bits that need updating
function rematch:SaveConfirm()
	rematch.savedTeams[confirmFrame.teamName] = {}
	local team = rematch.savedTeams[confirmFrame.teamName]
	for i=1,3 do
		team[i] = {}
		for j=1,4 do
			rematch.savedTeams[confirmFrame.teamName][i][j] = confirmFrame.newTeam[i][j]
		end
	end
	if confirmFrame.newTeam[4]~=0 then
		team[4] = confirmFrame.newTeam[4]
	end
	rematch.selectedTeam = confirmFrame.teamName
	rematch:UpdateSavedPets()
	for i=1,3 do
		frame.saved[i].shine:SetCooldown(0,1) -- do shiny cooldown ending bit
	end
	if frame.drawer:IsVisible() then
		rematch:OrderTeamList()
		rematch:UpdateTeamList()
	end
	incFrame:Hide()
	frame.share:Hide()
end

-- sets header to teamNpc and creates slots if they haven't already been defined
function rematch:PrepareConfirmation(teamName,npcID,delete)

	-- confirmFrame can be 258 height with secondTeam shown; 154 with secondTeam hidden
	confirmFrame.header.text:SetText(teamName)
	confirmFrame.header.text:SetTextColor(1,npcID and 1 or .82,npcID and 1 or 0)

	if delete then
		confirmFrame.question:SetText("Do you want to delete:")
		confirmFrame.secondTeam:Hide()
		confirmFrame:SetHeight(154)
	else
		confirmFrame.question:SetText("Do you want to replace:")
		confirmFrame.secondTeam:Show()
		confirmFrame:SetHeight(258)
	end

	if not confirmFrame.old then
		confirmFrame.old = {}
		rematch:CreateConfirmSlots(confirmFrame.old,confirmFrame,16,-54)
		confirmFrame.new = {}
		rematch:CreateConfirmSlots(confirmFrame.new,confirmFrame.secondTeam,16,-36)
		confirmFrame.secondTeam:SetFrameLevel(confirmFrame:GetFrameLevel())
	end
end

-- creates pet+3 ability buttons in passed table. incFrame also uses this
function rematch:CreateConfirmSlots(slotTable,parent,xoff,yoff)
	for i=1,3 do
		slotTable[i] = CreateFrame("Button",nil,parent,"RematchConfirmPetTemplate")
		local button = slotTable[i]
		button:SetPoint("TOPLEFT",(i-1)*78+xoff,yoff)
		button.ability = {}
		for j=1,3 do
			button.ability[j] = CreateFrame("Button",nil,button,"RematchSmallSlotTemplate")
			button.ability[j]:SetPoint("TOPLEFT",button,"TOPRIGHT",2,(j-1)*-22+10)
		end
	end
end

-- does a simple fill for pets+abilities based on team
function rematch:FillPetFrames(pets,team)
	rematch:WipePetFrames(pets)
	local petID,speciesID,icon,speciesName
	for i=1,3 do
		petID = team[i][1]
		if petID==0 then -- leveling slot
			icon = rematch.levelingIcon
		elseif type(petID)=="string" then
			icon = select(9,C_PetJournal.GetPetInfoByPetID(petID))
		elseif type(petID)=="number" then
			speciesName,icon = C_PetJournal.GetPetInfoBySpeciesID(petID)
		end
		if icon then
			pets[i].petID = petID
			pets[i].icon:SetTexture(icon)
			pets[i].icon:Show()
			if petID~=0 then
				for j=1,3 do
					pets[i].ability[j].abilityID = team[i][j+1]
					pets[i].ability[j].icon:SetTexture((select(3,C_PetBattles.GetAbilityInfoByID(team[i][j+1]))))
					pets[i].ability[j].icon:Show()
				end
			end
		end
	end
end

-- pops up a confirmation whether to delete a team (if auto confirm unchecked)
function rematch:DeleteTeam(teamName)
	confirmFrame.teamName = teamName
	if rematch.settings.AutoConfirm then
		rematch:DeleteConfirm()
	else
		rematch:PrepareConfirmation(teamName,rematch.savedTeams[teamName],true)
		rematch:FillPetFrames(confirmFrame.old,rematch.savedTeams[teamName])
		confirmFrame.func = rematch.DeleteConfirm
		confirmFrame:Show()
	end
end

-- does the actual delete
function rematch:DeleteConfirm()
	rematch.savedTeams[confirmFrame.teamName] = nil
	rematch.selectedTeam = nil
	rematch:UpdateSavedPets()
	if frame.drawer:IsVisible() then
		rematch:OrderTeamList()
		rematch:UpdateTeamList()
	end
end

--[[ Leveling Pet System ]]

-- pass no petID when just wanting to update saved pets
function rematch:SetLevelingPet(petID)
	local oldPetID = frame.drawer.leveling.petID
	if petID and C_PetJournal.GetPetInfoByPetID(petID) then
		rematch:DeleteFromLevelingQueue(petID) -- if petID already in queue, remove it
		tinsert(rematch.settings.LevelingQueue,1,petID) -- and insert it into top of queue
	end
	local currentPetID = rematch:GetLevelingPetID()
	frame.drawer.leveling.petID = currentPetID
	frame.drawer.leveling.icon:SetTexture(rematch:GetLevelingPetIcon())
	rematch:UpdateLevelingBorders()
	rematch:UpdateSavedPets()
	if oldPetID and currentPetID and oldPetID~=currentPetID then
		-- there was an old leveling pet and now there's a new one
		-- see if old one is in any loadout slots and replace with new one if so
		for i=1,3 do
			if C_PetJournal.GetPetLoadOutInfo(i)==oldPetID then
				rematch:SummonedPetMayChange()
				C_PetJournal.SetPetLoadOutInfo(i,currentPetID)
				break
			end
		end
	end
end

-- returns true if petID is below 25 and can battle
function rematch:PetCanLevel(petID)
	if petID or type(petID)=="string" then
		local _,_,level,_,_,_,_,_,_,_,_,_,_,_,canBattle = C_PetJournal.GetPetInfoByPetID(petID)
		if level and level<25 and canBattle then
			return true
		end
	end
end

-- returns true if the petID is the *current* leveling pet
function rematch:IsCurrentLevelingPet(petID)
	return petID and rematch.settings.LevelingQueue[1]==petID
end

-- returns index in queue and whether it's the last pet if petID is in queue
function rematch:PetIsLeveling(petID)
	if petID or type(petID)=="string" then
		local queue = rematch.settings.LevelingQueue
		for i=1,#queue do
			if queue[i]==petID then
				return i,i==#queue
			end
		end
	end
end


function rematch:UpdateLevelingBorders()
	local petID = rematch.settings.LevelingQueue[1]
	for i=1,3 do
		frame.current[i].leveling:SetShown(petID and frame.current[i].petID==petID)
	end
	frame.drawer.leveling.leveling:SetShown(petID~=nil)
end

function rematch:GetLevelingPetIcon()
	local petID = rematch.settings.LevelingQueue[1]
	local icon = petID and (select(9,C_PetJournal.GetPetInfoByPetID(petID))) or rematch.levelingIcon
	return icon
end

function rematch:GetLevelingPetID()
	return rematch.settings.LevelingQueue[1]
end

-- this is a prehook to the dropdown initialization for the journal pet menu (hook made when journal loads)
function rematch:NewPetMenuDropDownInit(level)
	local petID = PetJournal.menuPetID
	if petID then
	  local info = UIDropDownMenu_CreateInfo()
	  info.notCheckable = true
		if rematch:PetIsLeveling(petID) then
			info.text = "Stop Leveling"
			info.func = function() rematch:DeleteFromLevelingQueue(petID) rematch:ProcessLevelingQueue() end
		  UIDropDownMenu_AddButton(info,level)
		else
			if rematch:PetCanLevel(petID) then
				local info = UIDropDownMenu_CreateInfo()
				info.notCheckable = true
				info.text = "Start Leveling"
				info.func = function() rematch:SetLevelingPet(petID) rematch:ProcessLevelingQueue() end
				UIDropDownMenu_AddButton(info,level)
				if #rematch.settings.LevelingQueue>0 then
					info.text = "Add to Leveling Queue"
					info.func = function() rematch.dropDownPetID=petID rematch:DropDownFuncSendToEnd() end
					UIDropDownMenu_AddButton(info,level)
				end
			end
		end
	end
	rematch.oldDropDownInit(self,level)
end

function rematch:LevelingSlotOnEnter()
	if not queueFrame:IsVisible() then
		rematch:ProcessLevelingQueue()
		rematch:UpdateLevelingQueue()
		if #rematch.settings.LevelingQueue>0 then
			queueFrame:Show() -- show queue on mouseover (if there's already a leveling pet)
		end
	end
	rematch:ShowFloatingPetCard(self.petID,"BOTTOMLEFT",self,"BOTTOMRIGHT",0,0)
end

function rematch:LevelingSlotOnClick(button)
	if button=="RightButton" and rematch:GetLevelingPetID() then
		rematch.dropDownPetID = self.petID
		rematch:ShowDropDown(3)
	else
		rematch.PetOnClick(self,button)
	end
end

function rematch:DeleteFromLevelingQueue(petID)
	if petID then
		local queue = rematch.settings.LevelingQueue
		for i=1,#queue do
			if queue[i]==petID then
				table.remove(queue,i)
				return
			end
		end
	end
end

function rematch:ProcessLevelingQueue()
	if C_PetBattles.IsInBattle() then
		frame:RegisterEvent("PET_BATTLE_OVER") -- in case a pet levels to 25 when we're in pet battl
	end
	if InCombatLockdown() then
		rematch.queueNeedsProcessed = true -- in case a pet levels to 25 when we're in combat
		return
	end
	local queue = rematch.settings.LevelingQueue
	for i=#queue,1,-1 do -- remove any pets that no longer exist, are 25 or can't battle
		if not rematch:PetCanLevel(queue[i]) then
			table.remove(queue,i)
		end
	end
	rematch:SetLevelingPet()
end

-- updates the hybridscrollframe in the levelingqueue
function rematch:UpdateLevelingQueue()
	local queue = rematch.settings.LevelingQueue
	local numData = #queue
	local scrollFrame = queueFrame.scrollFrame
	local offset = HybridScrollFrame_GetOffset(scrollFrame)
	local buttons = scrollFrame.buttons
	local lastButton
	for i=1, #buttons do
		local index = i + offset
		local button = buttons[i]
		if ( index <= numData) then
			button.petID = queue[index]
			button.index = index
			local _,_,level,_,_,_,_,_,icon,petType = C_PetJournal.GetPetInfoByPetID(button.petID)
			if icon then
				button.icon:SetTexture(icon)
				button.level:SetText(level)
				local rarity = select(5,C_PetJournal.GetPetStats(button.petID))
		    local r,g,b = ITEM_QUALITY_COLORS[rarity-1].r, ITEM_QUALITY_COLORS[rarity-1].g, ITEM_QUALITY_COLORS[rarity-1].b
		    button.rarity:SetGradientAlpha("VERTICAL",r,g,b,.25,r,g,b,.6)
				if button==GetMouseFocus() then
					rematch.LevelingQueueOnEnter(button) -- we're scrolling, pretend moving into this button
				end
			end
			lastButton = button
			button:Show()
		else
			button.petID = nil
			button.index = nil
			button:Hide()
		end
	end
	-- a scrollframe can't receive clicks/drags; this invisible button will capture clicks/drags to empty area of scrollframe
	if numData<5 then
		queueFrame.emptyCapture:SetPoint("TOPLEFT",lastButton,"BOTTOMLEFT",0,0)
		queueFrame.emptyCapture:Show()
	else
		queueFrame.emptyCapture:Hide()
	end
	HybridScrollFrame_Update(scrollFrame,22*numData,22)
end

function rematch:LevelingQueueOnUpdate(elapsed)
	-- if mouse is no longer over queue/leveling slot and menu isn't up, the hide queue and process leveling queue in case anything moved to first
	if not queueFrame:IsMouseOver() and not frame.drawer.leveling:IsMouseOver() and not (UIDROPDOWNMENU_OPEN_MENU==rematch.dropDownFrame and DropDownList1:IsVisible()) then
		queueFrame.draggingPetID = nil
		queueFrame.draggingIndex = nil
		self:Hide()
		rematch:HideDropDown()
		rematch:ProcessLevelingQueue()
	else
		-- otherwise see if a pet is being dragged from within the queue and needs to swap
		local focus = GetMouseFocus()
		if queueFrame.draggingIndex and focus and focus.index and focus.index~=queueFrame.draggingIndex and focus.petID then
			local focusPetID = focus.petID
			rematch.settings.LevelingQueue[focus.index] = queueFrame.draggingPetID
			rematch.settings.LevelingQueue[queueFrame.draggingIndex] = focusPetID
			queueFrame.draggingIndex = focus.index
			rematch:UpdateLevelingQueue()
		end
	end
end

-- display floating pet card of pet within the queue
function rematch:LevelingQueueOnEnter()
	local button = self or GetMouseFocus()
	if button and button.petID then
		rematch:ShowFloatingPetCard(button.petID,"BOTTOMLEFT",button,"BOTTOMRIGHT",0,0)
	else
		rematch:HideFloatingPetCard()
	end
end

function rematch:LevelingQueueOnLeave()
	rematch:HideFloatingPetCard()
end

-- when pets within the queue are dragged around, note the ID and index for the OnUpdate to watch
function rematch:LevelingQueueOnDragStart()
	if not GetCursorInfo() then
		queueFrame.draggingPetID = self.petID
		queueFrame.draggingIndex = self.index
	end
end

-- when dragging stops, clear the dragging info; but note this handler happens anywhere the mouse button is released
function rematch:LevelingQueueOnDragStop()
	if frame.drawer.leveling:IsMouseOver() then -- if dragging to levelingslot, set the leveling pet
		rematch:SetLevelingPet(queueFrame.draggingPetID)
		rematch:UpdateLevelingQueue()
	end
	queueFrame.draggingPetID = nil
	queueFrame.draggingIndex = nil
end

-- when a pet within the queue is clicked
function rematch:LevelingQueueOnClick(button)
	if button=="RightButton" and self.petID then
		rematch.dropDownPetID = self.petID
		rematch:ShowDropDown(4)
	else
		if GetCursorInfo()=="battlepet" then
			rematch:LevelingQueueOnReceiveDrag()
		else
			rematch.PetOnClick(self,button) -- do left-click stuff
		end
	end
end

-- remember that this handler can trigger when draggingPetIDs are dragged around too
function rematch:LevelingQueueOnReceiveDrag()
	local cursorType,petID = GetCursorInfo()
	local focus = GetMouseFocus()
	if cursorType=="battlepet" and rematch:PetCanLevel(petID) then
		rematch:DeleteFromLevelingQueue(petID)
		if focus.index then -- dragged onto an existing pet in queue
			tinsert(rematch.settings.LevelingQueue,focus.index,petID)
		else -- dragged onto empty area of queue (add to end)
			tinsert(rematch.settings.LevelingQueue,petID)
		end
		rematch:ProcessLevelingQueue()
		rematch:UpdateLevelingQueue()
		ClearCursor()
	end
end

--[[ Pet DropDown Menus ]]

-- mode 1=current pet, 2=saved pets, 3=leveling slot, 4=leveling queue
function rematch:ShowDropDown(mode)
	rematch:HideFloatingPetCard()
	local menu = rematch.dropDownMenu
	wipe(menu)
	local petID = rematch.dropDownPetID
	if petID then
		local _,customName,_,_,_,_,_,realName = C_PetJournal.GetPetInfoByPetID(petID)
		tinsert(menu,{text=customName or realName,notCheckable=true,isTitle=true})
		if mode==3 or mode==4 then -- either leveling slot or leveling queue gets these options
			tinsert(menu,{text="Stop Leveling",notCheckable=true,func=rematch.DropDownFuncStopLeveling})
			local queueSlot,isLast = rematch:PetIsLeveling(petID)
			if queueSlot~=1 then
				tinsert(menu,{text="Send to Start of Queue",notCheckable=true,func=rematch.DropDownFuncSendToStart})
			end
			if not isLast then
				tinsert(menu,{text="Send to End of Queue",notCheckable=true,func=rematch.DropDownFuncSendToEnd})
			end
		elseif mode==1 then -- a current pet
			if rematch:PetCanLevel(petID) then
				if rematch:PetIsLeveling(petID) then
					tinsert(menu,{text="Stop Leveling",notCheckable=true,func=rematch.DropDownFuncStopLeveling})
				else
					tinsert(menu,{text="Start Leveling",notCheckable=true,func=rematch.DropDownFuncSendToStart})
				end
			end
		end

		-- dismiss/summon is common for all modes
		tinsert(menu,{text=C_PetJournal.GetSummonedPetGUID()==petID and PET_ACTION_DISMISS or SUMMON,notCheckable=true,func=rematch.DropDownFuncSummonOrDismiss})
	end

	tinsert(menu,{text="Cancel",notCheckable=true})
	EasyMenu(menu,rematch.dropDownFrame,"cursor",0,0,"MENU")
end

function rematch:HideDropDown()
	if UIDROPDOWNMENU_OPEN_MENU==rematch.dropDownFrame then
		CloseDropDownMenus()
	end
end

function rematch:DropDownFuncStopLeveling()
	rematch:DeleteFromLevelingQueue(rematch.dropDownPetID)
	rematch:ProcessLevelingQueue()
	rematch:UpdateLevelingQueue()
end

function rematch:DropDownFuncSendToStart()
	rematch:SetLevelingPet(rematch.dropDownPetID)
	rematch:UpdateLevelingQueue()
end

function rematch:DropDownFuncSendToEnd()
	rematch:DeleteFromLevelingQueue(rematch.dropDownPetID)
	tinsert(rematch.settings.LevelingQueue,rematch.dropDownPetID)
	rematch:ProcessLevelingQueue()
	rematch:UpdateLevelingQueue()
end

function rematch:DropDownFuncSummonOrDismiss()
	C_PetJournal.SummonPetByGUID(rematch.dropDownPetID)
end

-- does a "criterea" toast to alert that the leveling pet has changed
function rematch:ToastNextLevelingPet(petID)
	local _,customName,_,_,_,_,_,realName,icon = C_PetJournal.GetPetInfoByPetID(petID)
	if icon then
    local frame = CriteriaAlertFrame_GetAlertFrame()
		if frame then
			local frameName = frame:GetName()
			_G[frameName.."Name"]:SetText(customName or realName)
			_G[frameName.."Unlocked"]:SetText("Next leveling pet:")
	    _G[frameName.."IconTexture"]:SetTexture(icon)
	    frame.id = nil
	    AlertFrame_AnimateIn(frame)
      AlertFrame_FixAnchors()
			if not frame.rematchHide then
				-- add an OnHide secure hook to change "Unlocked" back to "Achievement Progress"
				frame.rematchHide = true
				frame:HookScript("OnHide",function(self) _G[self:GetName().."Unlocked"]:SetText(ACHIEVEMENT_PROGRESSED) end)
			end
		end
	end
end

--[[ Best of Species--aka Auto Upgrade option ]]

function rematch:SetupBestOfSpecies()
	if rematch.settings.BestOfSpecies then
		if not rematch.BoS then
			-- creating a separate frame to handle this that we can rip out when full journal list support added
			rematch.BoS = CreateFrame("Frame")
			rematch.BoS:SetScript("OnEvent",rematch.BestOfSpeciesOnEvent)
			rematch.BoS.filter = BATTLE_PET_NEW_PET:gsub("%%s","^(.-)")
			rematch.BoS.teamSpecies = {} -- will use to note species
		end
		rematch.BoS:RegisterEvent("CHAT_MSG_SYSTEM")
	elseif rematch.BoS then
		rematch.BoS:UnregisterEvent("CHAT_MSG_SYSTEM")
	end
end

function rematch:BestOfSpeciesOnEvent(event,msg)
	if event=="CHAT_MSG_SYSTEM" then -- when learning a pet, note its species and wait for it to enter journal
		local link = msg:match(rematch.BoS.filter)
		if link and link:match("battlepet") then
			rematch.BoS.speciesID = tonumber(link:match("battlepet:(%d+)"))
			rematch.BoS:RegisterEvent("PET_JOURNAL_LIST_UPDATE")
		end
	elseif event=="PET_JOURNAL_LIST_UPDATE" then -- should always be first fire after learning
		rematch.BoS:UnregisterEvent("PET_JOURNAL_LIST_UPDATE")

		-- get the best (highest level/rarity) petID for the learned species
		local bestPetID = rematch:GetBestPetIDForSpecies(rematch.BoS.speciesID)

		-- and if there is one, apply it to all teams that contain that species
		if bestPetID then
			for teamName,team in pairs(rematch.savedTeams) do
				for i=1,3 do
					local petID = team[i][1]
					-- if petID is the species we're upgrading, or the species of the petID is the species we're upgrading, set the bestPetID
					if (type(petID)=="number" and petID==rematch.BoS.speciesID) or (type(petID)=="string" and C_PetJournal.GetPetInfoByPetID(petID)==rematch.BoS.speciesID) then
						team[i][1] = bestPetID
					end
				end
			end
		end

		if frame:IsVisible() then
			rematch:UpdateSavedPets() -- update saved team display in case we have an affected team selected
		end
	end
end

-- go through the journal and get the highest level/rarity pet of a given species
function rematch:GetBestPetIDForSpecies(speciesID)
	local bestLevel,bestRarity,bestPetID = -1,-1
	for i=1,C_PetJournal.GetNumPets() do
		local petID,speciesID,_,_,level = C_PetJournal.GetPetInfoByIndex(i)
		if petID and speciesID==rematch.BoS.speciesID then
			rarity = select(5,C_PetJournal.GetPetStats(petID))
			if level>=bestLevel and rarity>=bestRarity then
				bestPetID = petID
				bestLevel = level
				bestRarity = rarity
			end
		end
	end
	return bestPetID
end
